# Full House -- A crazy queue system
# Copyright (C) 2007, 2008  Jaime Huerta-Cepas, Jordi Burguet-Castell

# This file is part of Full House.
#
# Full House is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Full House is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Full House.  If not, see <http://www.gnu.org/licenses/>.

# This module contains the common classes and functions for becario
# and ximo.

import sys, os
from logging import *

import odict


# --- COMMON CLASSES ---

class project:
    """Project consisting of multiple related jobs."""
    
    def __init__(self, name='', jobs=odict.OrderedDict(), uid=1000, environ={},
                 email='', nodes=[], priority=0, t_last=0,
                 totals={'W': 0,  'R': 0,  'S': 0,   'D': 0, 'E': 0, 'N': 0},
                 ti=0, tf=0, depends=[], status='R'):
        self.name = name
        
        self.jobs = jobs.copy()    # the keys are the alias of each job


        self.uid      = uid        # user id of the guy who sends the project
        self.environ  = environ.copy()  # all the environment variables
        self.email    = email      # email address to report when jobs finish
        self.nodes    = nodes[:]   # nodes where jobs are allowed to run in
        self.priority = priority   # priority of the project
        self.totals   = totals.copy()   # status totals
                                   # (Waiting Running Sleeping Done Error)
        self.ti       = ti         # starting time of the project
        self.tf       = tf         # finishing time of the project

        self.bcputime = {} # balanced (to other projs with same priority) cpu
                           # time spent per node (for sorting purposes)
        for node in self.nodes:
            self.bcputime[node] = 0
        
        self.t_last   = t_last     # time of last job sent from this project
                                   # (for sorting purposes)

        self.depends  = depends[:] # list of projects that must be
                                   # finished before this one starts

        self.status   = status     # general project status
                                   # Ready, Finished, Paused, Sleeping

        # Tal vez en el futuro cercano...
        #             maximum number of processes by node in gen,
        #             maximum number of processes in bioinfo
        #             maximum number of processes in ibms

    def update_totals(self):
        """Changes its status totals if needed."""

        # Do not go thru all the mess if we are done!
        if self.status == 'F':
            return
        
                      # Waiting  Running  Sleeping  Done    Error
        self.totals = {'W': 0,  'R': 0,  'S': 0,   'D': 0, 'E': 0, 'N': 0}
        for jname in self.jobs:
            job = self.jobs[jname]
            self.totals[job.status] += 1

        if self.totals['D'] + self.totals['E'] + self.totals['N'] == \
               len(self.jobs):
            self.status = 'F'


class job :
    """Each individual task the user wants to complete."""

    def __init__(self, jid='', command='', node='none', pid=0,
                 status='W', ti=0, tf=0):
        self.jid     = jid     # job id
        self.command = command

        self.node    = node    # node where process run
        self.pid     = pid

        self.status  = status
        self.ti      = ti      # starting time of the process
        self.tf      = tf      # finishing time of the process



# --- COMMON FUNCTIONS ---


def init_logging(appname, verbose=False, nodaemon=False):
    """Opens the file for logging stuff from application appname."""
    
    logger    = getLogger(appname)
    if nodaemon:  # if interactive, write output on stderr (screen)
        hdlr  = StreamHandler()
    else:         # if in daemon mode, write ouput to /var/tmp/app.log
        hdlr  = FileHandler('/var/tmp/' + appname + '.log')
    formatter = Formatter('%(asctime)s %(levelname)s %(message)s')
    hdlr.setFormatter(formatter)
    logger.addHandler(hdlr)
    if not verbose:
        logger.setLevel(WARNING)
    else:
        logger.setLevel(DEBUG)



def read_config(filename='/scheduler/nodes.conf'):
    """Reads the config file and returns the frontend, nodes and its
    maximum load and number of jobs."""
    
    frontend = ''
    nodes    = []
    maxjobs  = {}
    maxload  = {}
    for line in open(filename):
        if len(line) < 2 or line[0] == '#':  # skip void lines and comments
            continue
        if not frontend:    # the first to be read is the frontend
            if ' ' in line:
                print 'Error in config file: frontend must be a single name.'
                sys.exit(1)
            frontend = line
        else:
            try:
                name, mjobs, mload = line.split()
                nodes.append(name)
                maxjobs[name] = int(mjobs)
                maxload[name] = float(mload)
            except ValueError, error:
                print error
                sys.exit()
    return frontend, nodes, maxjobs, maxload



# Act like a daemon (see the "Unix Programming FAQ"
# http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16 )
def daemonize():
    """Become a daemon: fork, decouple from parent environment,
    fork again, close all files and reopen on /dev/null."""

    # Do first fork.
    try:
        pid = os.fork()
        if pid > 0:
            os._exit(0)  # exit first parent
    except OSError, e: 
        print >>sys.stderr, "Fork #1 failed: (%d) %s" % (e.errno, e.strerror)
        sys.exit(1) 
        
    # Decouple from parent environment. 
    os.chdir("/") 
    os.umask(0) 
    os.setsid() 
 
    # Do second fork.
    try:
        pid = os.fork()
        if pid > 0:
            os._exit(0)  # exit second parent
    except OSError, e:
        print >>sys.stderr, "Fork #2 failed: (%d) %s" % (e.errno, e.strerror)
        sys.exit(1) 
 
    # Now I am a daemon! 
 
    # Redirect standard file descriptors.
    sys.stdin.close()
    sys.stdout.close()
    sys.stderr.close()
    
    for i in xrange(1023):
        try:
            os.close(i)
        except OSError:
            pass
        
    sys.stdin  = open('/dev/null', 'r')   # fd 0
    sys.stdout = open('/dev/null', 'a+')  # fd 1
    sys.stderr = open('/dev/null', 'a+')  # fd 2
